__all__ = ('Button', 'ButtonStyle', 'Row')

import reprlib

from ...backend.utils import copy_docs

from ...discord.preinstanced import ButtonStyle, ComponentType
from ...discord.interaction import ComponentBase, Component
from ...discord.preconverters import preconvert_preinstanced_type
from ...discord.emoji import create_partial_emoji_data, create_partial_emoji, Emoji
from ...discord.limits import COMPONENT_SUB_COMPONENT_LIMIT, COMPONENT_LABEL_LENGTH_MAX, COMPONENT_CUSTOM_ID_LENGTH_MAX


COMPONENT_TYPE_ACTION_ROW = ComponentType.action_row
COMPONENT_TYPE_BUTTON = ComponentType.button


class Button(ComponentBase):
    """
    Button message component.
    
    Attributes
    ----------
    _data : `dict` of (`str`, `Any`) items
        Serialized component data.
    
    Class Attributes
    ----------------
    type : ``ComponentType`` = `ComponentType.button`
        The component's type.
    default_style : ``ButtonStyle`` = `ButtonStyle.secondary`
        The default button style to use if style is not given.
    """
    type = COMPONENT_TYPE_BUTTON
    default_style = ButtonStyle.secondary
    
    __slots__ = ('_data', )
    
    def __new__(cls, label=None, emoji=None, *, custom_id=None, url=None, style=None, enabled=True):
        """
        Creates a new component instance with the given parameters.
        
        Parameters
        ----------
        label : `None` or `str`
            Label of the component.
        emoji : `None` or ``Emoji``
            Emoji of the button if applicable.
        custom_id : `None` or `str`, Optional (Keyword only)
            Custom identifier to detect which button was clicked by the user.
            
            > Mutually exclusive with the `url` field.

        url : `None` or `str`, Optional (Keyword only)
            Url to redirect to when clicking on the button.
            
            > Mutually exclusive with the `custom_id` field.
        
        style : ``ButtonStyle``, `int`, Optional (Keyword only)
            The components's style. Applicable for buttons.
        
        enabled : `bool`, Optional (Keyword only)
            Whether the button is enabled. Defaults to `True`.
        
        Raises
        ------
        TypeError
            - If `style`'s type is unexpected.
        AssertionError
            - If `custom_id` was not given neither as `None` or `str` instance.
            - `url` is mutually exclusive with `custom_id`.
            - If `emoji` was not given as ``Emoji`` instance.
            - If `url` was not given neither as `None` or `str` instance.
            - If `style` was not given as any of the `type`'s expected styles.
            - If `label` was not given neither as `None` nor as `int` instance.
            - If `enabled` was not given as `bool` instance.
            - If `label`'s length is over `80`.
            - If `custom_id`'s length is over `100`.
        """
        if style is None:
            style = cls.default_style
        else:
            style = preconvert_preinstanced_type(style, 'style', ButtonStyle)
        
        data = {
            'type': cls.type.value,
            'style': style.value,
        }
        
        if __debug__:
            if (custom_id is not None) and (url is not None):
                raise AssertionError(f'`custom_id` and `url` fields are mutually exclusive, got '
                    f'custom_id={custom_id!r}, url={url!r}.')
        
        if (custom_id is not None):
            if __debug__:
                if (not isinstance(custom_id, str)):
                    raise TypeError(f'`custom_id` can be given either as `None` or as `str` instance, got '
                        f'{custom_id.__class__.__name__}.')
            
            data['custom_id'] = custom_id
        
        if (emoji is not None):
            if __debug__:
                if (not isinstance(emoji, Emoji)):
                    raise AssertionError(f'`emoji` can be given as `{Emoji.__name__}` instance, got '
                        f'{emoji.__class__.__name__}')
            
            data['emoji'] = create_partial_emoji_data(emoji)
        
        if (url is not None):
            if __debug__:
                if (not isinstance(url, str)):
                    raise AssertionError(f'`url` can be given either as `None` or as `str` instance, got '
                        f'{url.__class__.__name__}.')
            
            data['url'] = url
            
        if (label is not None):
            if __debug__:
                if (not isinstance(label, str)):
                    raise AssertionError(f'`label` can be given either as `None` or as `str` instance, got '
                        f'{label.__class__.__name__}.')
                
                if len(label) > COMPONENT_LABEL_LENGTH_MAX:
                    raise AssertionError(f'`label`\'s max length can be {COMPONENT_LABEL_LENGTH_MAX!r}, got '
                        f'{len(label)!r}; {label!r}.')
            
            if label:
                data['label'] = label
        
        if __debug__:
            if not isinstance(enabled, bool):
                raise AssertionError(f'`enabled` can be given as `bool` instance, got {enabled.__class__.__name__}.')
        
        if not enabled:
            data['disabled'] = True
        
        self = object.__new__(cls)
        self._data = data
        return self
    
    
    @property
    def style(self):
        """
        A get-set property for accessing the button's `style`.
        
        Accepts and returns ``ButtonStyle`` and `int` instances.
        """
        return ButtonStyle.get(self._data['style'])
    
    @style.setter
    def style(self, style):
        style = preconvert_preinstanced_type(style, 'style', ButtonStyle)
        self._data['style'] = style
    
    
    @property
    def custom_id(self):
        """
        A get-set property for accessing the button's `custom_id`.
        
        Accepts and returns ``None`` and `str` instances.
        """
        return self._data.get('custom_id', None)
    
    @custom_id.setter
    def custom_id(self, custom_id):
        if (custom_id is None):
            try:
                del self._data['custom_id']
            except KeyError:
                pass
        else:
            if __debug__:
                if (not isinstance(custom_id, str)):
                    raise TypeError(f'`custom_id` can be given either as `None` or as `str` instance, got '
                        f'{custom_id.__class__.__name__}.')
                
                if len(custom_id) > COMPONENT_CUSTOM_ID_LENGTH_MAX:
                    raise AssertionError(f'`custom_id`\'s max length can be {COMPONENT_CUSTOM_ID_LENGTH_MAX!r}, got '
                        f'{len(custom_id)!r}; {custom_id!r}.')
            
            self._data['custom_id'] = custom_id
    
    
    @property
    def emoji(self):
        """
        A get-set property for accessing the button's `emoji`.
        
        Accepts and returns ``None`` and ``Emoji`` instances.
        """
        try:
            emoji_data = self._data['emoji']
        except KeyError:
            emoji = None
        else:
            emoji = create_partial_emoji(emoji_data)
        
        return emoji
    
    @emoji.setter
    def emoji(self, emoji):
        if (emoji is None):
            try:
                del self._data['emoji']
            except KeyError:
                pass
        else:
            if __debug__:
                if (not isinstance(emoji, Emoji)):
                    raise TypeError(f'`emoji` can be given either as `None` or as `{Emoji.__name__}` instance, got '
                        f'{Emoji.__class__.__name__}.')
            
            self._data['Emoji'] = create_partial_emoji_data(emoji)
    
    
    @property
    def url(self):
        """
        A get-set property for accessing the button's `url`.
        
        Accepts and returns ``None`` and `str` instances.
        """
        return self._data.get('url', None)
    
    @url.setter
    def url(self, url):
        if (url is None):
            try:
                del self._data['url']
            except KeyError:
                pass
        else:
            if __debug__:
                if (not isinstance(url, str)):
                    raise TypeError(f'`url` can be given either as `None` or as `str` instance, got '
                        f'{url.__class__.__name__}.')
            
            self._data['url'] = url
    
    
    @property
    def label(self):
        """
        A get-set property for accessing the button's `label`.
        
        Accepts and returns ``None`` and `str` instances.
        """
        return self._data.get('url', None)
    
    @label.setter
    def label(self, label):
        if (label is not None):
            if __debug__:
                if (not isinstance(url, str)):
                    raise TypeError(f'`label` can be given either as `None` or as `str` instance, got '
                        f'{label.__class__.__name__}.')
                
                if len(label) > COMPONENT_LABEL_LENGTH_MAX:
                    raise AssertionError(f'`label`\'s max length can be {COMPONENT_LABEL_LENGTH_MAX!r}, got '
                        f'{len(label)!r}; {label!r}.')
            
            if label:
                self._data['label'] = label
                return
        
        try:
            del self._data['label']
        except KeyError:
            pass
    
    
    @property
    def enabled(self):
        """
        A get-set property for accessing the button's `enabled` field.
        
        Accepts and returns `None` and `str` instance.
        """
        return (not self._data.get('disabled', False))
    
    @enabled.setter
    def enabled(self, value):
        if __debug__:
            if not isinstance(enabled, bool):
                raise AssertionError(f'`enabled` can be given as `bool` instance, got {enabled.__class__.__name__}.')
        
        if value:
            try:
                del self._data['disabled']
            except KeyError:
                pass
        else:
            self._data['disabled'] = True
    
    
    @classmethod
    @copy_docs(ComponentBase.from_data)
    def from_data(cls, data):
        if __debug__:
            if data['type'] != cls.type.value:
                raise AssertionError(f'The received data is not a {cls.type.name} component, got: {data!r}.')
        
        self = object.__new__(cls)
        self._data = data
        return self
    
    
    @copy_docs(ComponentBase.to_data)
    def to_data(self):
        return self._data
    
    
    @copy_docs(ComponentBase.__repr__)
    def __repr__(self):
        repr_parts = [self.__class__.__name__, '(']
        
        data = self._data
        try:
            label = data['label']
        except KeyError:
            added_field = False
        else:
            repr_parts.append(reprlib.repr(label))
            added_field = True
        
        try:
            emoji = data['emoji']
        except KeyError:
            pass
        else:
            if added_field:
                repr_parts.append(', ')
            else:
                added_field = True
                repr_parts.append('emoji=')
            
            repr_parts.append(repr(emoji))
        
        for field_name in ('custom_id', 'url'):
            try:
                field_value = data[field_name]
            except KeyError:
                pass
            else:
                if added_field:
                    repr_parts.append(', ')
                else:
                    added_field = True
                
                repr_parts.append(field_name)
                repr_parts.append('=')
                repr_parts.append(reprlib.repr(field_value))
        
        style = data['style']
        if style != self.default_style:
            if added_field:
                repr_parts.append(', ')
            else:
                added_field = True
            repr_parts.append('style=')
            repr_parts.append(repr(style))
        
        
        enabled = not data.get('disabled', False)
        if (not enabled):
            if added_field:
                repr_parts.append(', ')
            
            repr_parts.append('enabled=')
            repr_parts.append(repr(enabled))
        
        repr_parts.append(')')
        return ''.join(repr_parts)
    
    
    @copy_docs(ComponentBase.copy)
    def copy(self):
        new = object.__new__(type(self))
        new._data = self._data.copy()
        return new
    
    
    @copy_docs(ComponentBase.__eq__)
    def __eq__(self, other):
        if type(other) is type(self):
            if self._data == other._data:
                return True
            else:
                return False
        
        if isinstance(other, ComponentBase):
            if self.type is not other.type:
                return False
            
            if self.emoji is not other.emoji:
                return False
            
            if self.style is not other.style:
                return False
            
            if self.custom_id != other.custom_id:
                return False
            
            if self.label != other.label:
                return False
            
            if self.enabled != other.enabled:
                return False
            
            return True
        
        
        return NotImplemented
    
    
    @copy_docs(ComponentBase.__hash__)
    def __hash__(self):
        data = self._data
        hash_value = self.type.value
        
        try:
            emoji = data['emoji']
        except KeyError:
            pass
        else:
            hash_value ^= emoji.id
        
        hash_value ^= data['style'].value
        
        for field_name in ('custom_id', 'url', 'label'):
            try:
                field_value = data[field_name]
            except KeyError:
                pass
            else:
                hash_value ^= hash(field_value)
        
        enabled = not data.get('disabled', False)
        if enabled:
            hash_value ^= 1<<8
        
        return hash_value


class Row(ComponentBase):
    """
    Action row message component.
    
    Attributes
    ----------
    _data : `dict` of (`str`, `Any`) items
        Serialized component data.
    
    Class Attributes
    ----------------
    type : ``ComponentType`` = `ComponentType.action_row`
        The component's type.
    """
    type = COMPONENT_TYPE_ACTION_ROW
    
    __slots__ = ('_data', )
    
    def __new__(cls, *components):
        component_datas = None
        for component in components:
            if __debug__:
                if not isinstance(component, ComponentBase):
                    raise AssertionError(f'`component` can be given as `{ComponentBase.__name__}` instance, got '
                        f'{component.__class__.__name__}.')
                
                if component.type is COMPONENT_TYPE_ACTION_ROW:
                    raise AssertionError(f'Cannot add `{COMPONENT_TYPE_ACTION_ROW}` type as sub components, got '
                        f'{component!r}.')
            
            if component_datas is None:
                component_datas = []
            
            component_datas.append(component.to_data())
        
        if __debug__:
            if (component_datas is not None) and (len(component_datas) > COMPONENT_SUB_COMPONENT_LIMIT):
                raise AssertionError(f'A `{cls.__name__}` can have maximum 5 sub-components, got '
                    f'{len(component_datas)}; {component_datas!r}.')
        
        data = {
            'type': cls.type.value
        }
        
        if (components is not None):
            data['components'] = component_datas
        
        self = object.__new__(cls)
        self._data = data
        return self
    
    
    @property
    def components(self):
        """
        A get property for accessing the row's components.
        
        Accepts `None` or a `list` of ``ComponentBase`` instances.
        """
        try:
            component_datas = self._data['components']
        except KeyError:
            components = None
        else:
            components = [
                COMPONENT_TYPE_MAP.get(component_data['type'], Component).from_data(component_data)
                    for component_data in component_datas
            ]
        
        return components


    @classmethod
    @copy_docs(ComponentBase.from_data)
    def from_data(cls, data):
        if __debug__:
            if data['type'] != cls.type.value:
                raise AssertionError(f'The received data is not a {cls.type.name} component, got: {data!r}.')
        
        self = object.__new__(cls)
        self._data = data
        return self
    
    
    @copy_docs(ComponentBase.to_data)
    def to_data(self):
        return self._data
    
    
    @copy_docs(ComponentBase.__repr__)
    def __repr__(self):
        repr_parts = [self.__class__.__name__, '(']
        
        try:
            component_datas = self._data['components']
        except KeyError:
            pass
        else:
            field_added = False
            for component_data in component_datas:
                if field_added:
                    repr_parts.append(', ')
                else:
                    field_added = True
                
                component = COMPONENT_TYPE_MAP.get(component_data['type'], Component).from_data(component_data)
                repr_parts.append(component)
                
        repr_parts.append(')')
        return ''.join(repr_parts)
    
    
    @copy_docs(ComponentBase.copy)
    def copy(self):
        old_data = self._data
        new_data = {'type': old_data['style']}
        try:
            component_datas = old_data['components']
        except KeyError:
            pass
        else:
            new_data['components'] = [component_data.copy() for component_data in component_datas]
        
        new = object.__new__(type(self))
        new._data = new_data
        return new


    @copy_docs(ComponentBase.__eq__)
    def __eq__(self, other):
        if type(other) is  type(self):
            if self._data == other._data:
                return True
            else:
                return False
        
        if isinstance(other, ComponentBase):
            if self.type is not other.type:
                return False
            
            if self.components != other.components:
                return False
            
            return True
        
        return NotImplemented
    
    
    @copy_docs(ComponentBase.__hash__)
    def __hash__(self):
        hash_value = self.type.value
        
        components = self.components
        if (components is not None):
            hash_value ^= len(components)<<12
            for component in components:
                hash_value ^= hash(component)
        
        return hash_value


        
COMPONENT_TYPE_MAP = {
    COMPONENT_TYPE_ACTION_ROW.value: Row,
    COMPONENT_TYPE_BUTTON.value: Button,
}





# -*- coding: utf-8 -*-
__all__ = ('DATETIME_FORMAT_CODE', 'DISCORD_EPOCH', 'Gift', 'Relationship', 'Unknown', 'cchunkify', 'chunkify',
    'elapsed_time', 'filter_content', 'id_to_time', 'is_id', 'is_invite_code', 'is_mention', 'is_role_mention',
    'is_user_mention', 'now_as_id', 'parse_message_reference', 'parse_rdelta', 'parse_tdelta', 'random_id',
    'sanitize_mentions', 'time_to_id',)

import random, sys
from re import compile as re_compile
from datetime import datetime, timedelta
from base64 import b64encode
from time import time as time_now


try:
    from dateutil.relativedelta import relativedelta
except ImportError:
    relativedelta = None

from ..backend.export import export, include
from ..backend.utils import istr, modulize

from .bases import DiscordEntity
from .client_core import USERS, CHANNELS, ROLES

MESSAGE_JUMP_URL_RP = include('MESSAGE_JUMP_URL_RP')
create_partial_user = include('create_partial_user')
RelationshipType = include('RelationshipType')

DATETIME_FORMAT_CODE = '%Y-%m-%d %H:%M:%S'

def endswith_xFFxD9(data):
    """
    Checks whether the given data endswith `b'\xD9\xFF'` ignoring empty bytes at the end of it.
    
    Parameters
    ----------
    data : `bytes-like`
    
    Returns
    -------
    result : `bool`
    """
    index = len(data)-1
    while index > 1:
        actual = data[index]
        if actual == b'\xD9'[0] and data[index-1] == b'\xFF'[0]:
            return True
        
        if actual:
            return False
        
        index -= 1
        continue

def get_image_extension(data):
    """
    Gets the given raw image data's extension and returns it.
    
    Parameters
    ----------
    data : `bytes-like`
        Image data.
    
    Returns
    -------
    extension_name : `str`
    """
    if data.startswith(b'\x89\x50\x4E\x47\x0D\x0A\x1A\x0A'):
        extension_name = 'png'
    elif data.startswith(b'\xFF\xD8') and endswith_xFFxD9(data):
        extension_name = 'jpeg'
    elif data.startswith(b'\x47\x49\x46\x38\x37\x61') or data.startswith(b'\x47\x49\x46\x38\x39\x61'):
        extension_name = 'gif'
    else:
        extension_name = ''
    
    return extension_name

def image_to_base64(data):
    """
    Converts a bytes image to a base64 one.
    
    Parameter
    ----------
    data : `bytes-like`
        Image data.
    
    Returns
    -------
    base64 : `str`
    
    Raises
    ------
    ValueError
        If `ext` was not given and the given `data`'s image format is not any of the expected ones.
    """
    if data.startswith(b'\x89\x50\x4E\x47\x0D\x0A\x1A\x0A'):
        extension_value = 'image/png'
    elif data.startswith(b'\xFF\xD8') and endswith_xFFxD9(data):
        extension_value = 'image/jpeg'
    elif data.startswith(b'\x47\x49\x46\x38\x37\x61') or data.startswith(b'\x47\x49\x46\x38\x39\x61'):
        extension_value = 'image/gif'
    else:
        raise ValueError('Unsupported image type given.')
    
    return ''.join(['data:', extension_value, ';base64,', b64encode(data).decode('ascii')])

DISCORD_EPOCH = 1420070400000
# example dates:
# "2016-03-31T19:15:39.954000+00:00"
# "2019-04-28T15:14:38+00:00"
# at edit:
# "2019-07-17T18:52:50.758993+00:00" #this is before desuppress!
# at desuppress:
# "2019-07-17T18:52:50.758000+00:00"

PARSE_TIME_RP = re_compile('(\\d{4})-(\\d{2})-(\\d{2})T(\\d{2}):(\\d{2}):(\\d{2})(?:\\.(\\d{3})?)?.*')

def parse_time(timestamp):
    """
    Parses the given timestamp.
    
    The expected timestamp formats are the following:
        - `2019-04-28T15:14:38+00:00`
        - `2019-07-17T18:52:50.758993+00:00`
        - `2019-07-17T18:52:50.758000+00:00`
        
    Discord might give timestamps with different accuracy, so we use an optimal middle way at parsing them.
    Some events depend on timestamp accuracy, so we really do not want to be wrong about them, or it might cause
    same internal derpage.
    
    If parsing a timestamp failed, the start of the discord epoch is returned and an error message is given at
    `sys.stderr`.
    
    Parameters
    ----------
    timestamp : `str`
        The timestamp to parse.
    
    Returns
    -------
    time : `datetime`
    
    Notes
    -----
    I already noted that timestamp formats are inconsistent, but even our baka Chiruno could have fix it...
    """
    parsed = PARSE_TIME_RP.fullmatch(timestamp)
    if parsed is None:
        sys.stderr.write(f'Cannot parse timestamp: `{timestamp}`, returning `DISCORD_EPOCH_START`\n')
        return DISCORD_EPOCH_START
    
    year = int(parsed.group(1))
    month = int(parsed.group(2))
    day = int(parsed.group(3))
    hour = int(parsed.group(4))
    minute = int(parsed.group(5))
    second = int(parsed.group(6))
    micro = parsed.group(7)
    
    if micro is None:
        micro = 0
    else:
        micro = int(micro)
    
    return datetime(year, month, day, hour, minute, second, micro)

@export
def id_to_time(id_):
    """
    Converts the given id to datetime.
    
    Parameters
    ----------
    id_ : `int`
        Unique identifier number of a Discord entity.
    
    Returns
    -------
    time : `datetime`
    """
    return datetime.utcfromtimestamp(((id_>>22)+DISCORD_EPOCH)/1000.)

DISCORD_EPOCH_START = id_to_time(0)

def time_to_id(time):
    """
    Converts the given time to it's respective discord identifier number.
    
    Parameters
    ----------
    time : `datetime`
    
    Returns
    -------
    id_ `int`
    """
    return ((time.timestamp()*1000.).__int__()-DISCORD_EPOCH)<<22

def random_id():
    """
    Generates a random Discord identifier number what's datetime value did not surpass the current time.
    
    Returns
    -------
    id_ `int`
    """
    return (((time_now()*1000.).__int__()-DISCORD_EPOCH)<<22)+(random.random()*4194304.).__int__()


def log_time_converter(value):
    """
    Converts the given value to it's snowflake representation.
    
    Parameters
    ----------
    value : `int`, ``DiscordEntity`` instance or `datetime`
        If the value is given as `int`, returns it. If given as a ``DiscordEntity``, then returns it's id and if it
        is given as a `datetime` object, then converts that to snowflake then returns it.
    
    Returns
    -------
    snowflake : `int`
    
    Raises
    ------
    TypeError
        If the given `value`'s type is not any of the expected ones.
    """
    if isinstance(value, int):
        return value
    
    if isinstance(value, DiscordEntity):
        return value.id
    
    if type(value) is datetime:
        return time_to_id(value)
    
    raise TypeError(f'Expected `int`, `{DiscordEntity.__name__}` instance, or a `datetime` object, got '
        f'`{value.__class__.__name__}`.')

APPLICATION_COMMAND_NAME_RP = re_compile('[a-zA-Z0-9_\-]{1,32}')

ID_RP = re_compile('(\d{7,21})')
IS_MENTION_RP = re_compile('@(?:everyone|here)|<(?:@[!&]?|#|/[a-zA-Z0-9_\-]{3,32}:)\d{7,21}>')

USER_MENTION_RP = re_compile('<@!?(\d{7,21})>')
CHANNEL_MENTION_RP = re_compile('<#(\d{7,21})>')
ROLE_MENTION_RP = re_compile('<@&(\d{7,21})>')
APPLICATION_COMMAND_MENTION_RP = re_compile('</([a-zA-Z0-9_\-]{3,32}):(\d{7,21})>')

EMOJI_RP = re_compile('<(a)?:([a-zA-Z0-9_]{2,32})(?:~[1-9])?:(\d{7,21})>')
EMOJI_NAME_RP = re_compile(':?([a-zA-Z0-9_\\-~]{1,32}):?')
FILTER_RP = re_compile('("(.+?)"|\S+)')
INVITE_CODE_RP = re_compile('([a-zA-Z0-9-]+)')

def is_valid_application_command_name(name):
    """
    Returns whether the given application command name is valid.
    
    Parameters
    ----------
    name : `str`
        The name of the application command.
    
    Returns
    -------
    valid : `bool`
    """
    return (APPLICATION_COMMAND_NAME_RP.fullmatch(name) is not None)

def is_id(text):
    """
    Returns whether the given text is a valid snowflake.
    
    Parameters
    ----------
    text : `str`
    
    Returns
    -------
    result : `bool`
    """
    return ID_RP.fullmatch(text) is not None

def is_mention(text):
    """
    Returns whether the given text is a mention.
    
    Parameters
    ----------
    text : `str`
    
    Returns
    -------
    result : `bool`
    """
    return IS_MENTION_RP.fullmatch(text) is not None

def is_user_mention(text):
    """
    Returns whether the given text mentions a user.
    
    Parameters
    ----------
    text : `str`
    
    Returns
    -------
    result : `str`
    """
    return USER_MENTION_RP.fullmatch(text) is not None

def is_channel_mention(text):
    """
    Returns whether the given text mentions a channel.
    
    Parameters
    ----------
    text : `str`
    
    Returns
    -------
    result : `bool`
    """
    return CHANNEL_MENTION_RP.fullmatch(text) is not None

def is_role_mention(text):
    """
    Returns whether the given text mentions a role.
    
    Parameters
    ----------
    text : `str`
    
    Returns
    -------
    result : `bool`
    """
    return ROLE_MENTION_RP.fullmatch(text) is not None

def is_application_command_mention(text):
    """
    Returns whether the given text is an interaction command mention.
    
    Parameters
    ----------
    text : `str`
    
    Returns
    -------
    result : `bool`
    """
    return (APPLICATION_COMMAND_MENTION_RP.fullmatch(text) is not None)

def is_invite_code(text):
    """
    Returns whether the given text is an invite code
    
    Parameters
    ----------
    text : `str`
    
    Returns
    -------
    result : `bool`
    """
    return (INVITE_CODE_RP.fullmatch(text) is not None)

def now_as_id():
    """
    Returns the current time as a Discord snowflake.
    
    Returns
    -------
    snowflake : `int`
    """
    return ((time_now()*1000.)-DISCORD_EPOCH).__int__()<<22

def filter_content(content):
    """
    Filters the given content to parts separated with spaces. Parts surrounded with `"` character will count as one
    even if they contain spaces.
    
    Parameters
    ----------
    content : `str`
    
    Returns
    -------
    parts : `list` of `str`
    """
    return [match[1] or match[0] for match in FILTER_RP.findall(content)]

def chunkify(lines, limit=2000):
    """
    Creates chunks of strings from the given lines.
    
    Parameters
    ----------
    lines : `list` of `str`
        Lines of text to be chunkified.
    limit : `int`, Optional
        The maximal length of a generated chunk.
    
    Returns
    -------
    result : `list` of `str`
    
    Raises
    ------
    ValueError`
        If limit is less than `500`.
    """
    if limit < 500:
        raise ValueError(f'Minimal limit should be at least 500, got {limit!r}.')
    
    result = []
    chunk_ln = 0
    chunk = []
    for line in lines:
        while True:
            ln = len(line)+1
            if chunk_ln+ln > limit:
                position = limit-chunk_ln
                if position < 250:
                    result.append('\n'.join(chunk))
                    chunk.clear()
                    chunk.append(line)
                    chunk_ln=ln
                    break
                
                position = line.rfind(' ', position-250, position-3)
                if position == -1:
                    position = limit-chunk_ln-3
                    post_part = line[position:]
                else:
                    post_part = line[position+1:]
                
                pre_part = line[:position]+'...'
                
                chunk.append(pre_part)
                result.append('\n'.join(chunk))
                chunk.clear()
                if len(post_part) > limit:
                    line = post_part
                    chunk_ln = 0
                    continue
                
                chunk.append(post_part)
                chunk_ln = len(post_part)+1
                break
            
            chunk.append(line)
            chunk_ln += ln
            break
    
    result.append('\n'.join(chunk))
    
    return result

def cchunkify(lines, lang='', limit=2000):
    """
    Creates code block chunks from the given lines.
    
    Parameters
    ----------
    lines : `list` of `str`
        Lines of text to be chunkified.
    lang : `str`, Optional
        Language prefix of the code-block.
    limit : `int`, Optional
        The maximal length of a generated chunk.
    
    Returns
    -------
    result : `list` of `str`
    
    Raises
    ------
    ValueError`
        If limit is less than `500`.
    """
    if limit < 500:
        raise ValueError(f'Minimal limit should be at least 500, got {limit!r}.')
    
    starter = f'```{lang}'
    limit = limit-len(starter)-5
    
    result = []
    chunk_ln = 0
    chunk = [starter]
    for line in lines:
        while True:
            ln = len(line)+1
            if chunk_ln+ln > limit:
                position = limit-chunk_ln
                if position < 250:
                    chunk.append('```')
                    result.append('\n'.join(chunk))
                    chunk.clear()
                    chunk.append(starter)
                    chunk.append(line)
                    chunk_ln = ln
                    break
                
                position = line.rfind(' ', position-250, position-3)
                if position == -1:
                    position = limit-chunk_ln-3
                    post_part = line[position:]
                else:
                    post_part = line[position+1:]
                
                pre_part = line[:position]+'...'
                
                chunk.append(pre_part)
                chunk.append('```')
                result.append('\n'.join(chunk))
                chunk.clear()
                chunk.append(starter)
                
                if len(post_part) > limit:
                    line = post_part
                    chunk_ln = 0
                    continue
                
                chunk.append(post_part)
                chunk_ln = len(post_part)+1
                break
            
            chunk.append(line)
            chunk_ln += ln
            break
    
    if len(chunk)>1:
        chunk.append('```')
        result.append('\n'.join(chunk))
    
    return result

if relativedelta is None:
    elapsed_time = None
else:
    def elapsed_time(delta, limit=3, names=(
            ('year', 'years',),
            ('month', 'months'),
            ('day', 'days', ),
            ('hour', 'hours'),
            ('minute', 'minutes'),
            ('second', 'seconds'),)):
        """
        Generates an elapsed time formula from the given time delta.
        
        Parameters
        ----------
        delta : `datetime` or `relativedelta`
            The time delta. If given as `datetime`, then the delta will be based on the difference between the given
            datetime and the actual time. If given as `relativedelta`, then that will be used directly.
        limit : `int`, Optional
            The maximal amount of connected time units. Defaults to `3`.
        names : `iterable` of `tuple` (`str`, `str`), Optional
            The names of the time units starting from years. Each element of the iterable should yield a `tuple` of two
            `str` elements. The first should be always the singular form of the time unit's name and the second the
            plural. Defaults to the time units' names in engrisssh.
        
        Returns
        -------
        result : `str`
        
        Raises
        ------
        TypeError
            If delta was neither passed as `datetime` or as `relativedelta` instance.
        """
        if isinstance(delta, datetime):
            delta = relativedelta(datetime.utcnow(), delta)
        elif isinstance(delta, relativedelta):
            pass
        else:
            raise TypeError(f'Expected, `relativedelta` or `datetime`, got {delta.__class__.__name__}.')
        
        parts = []
        for value, name_pair in zip((delta.years, delta.months, delta.days, delta.hours, delta.minutes, delta.seconds), names):
            if limit == 0:
                break
            
            if value < 0:
                 value = -value
            elif value == 0:
                continue
            
            parts.append(str(value))
            parts.append(' ')
            
            name = name_pair[value!=1]
            
            parts.append(name)
            parts.append(', ')
            
            limit -= 1
        
        if parts:
            del parts[-1]
            result = ''.join(parts)
        else:
            result = f'0 {names[5][0]}'
        
        return result


class Relationship:
    """
    Represents a Discord relationship.
    
    Attributes
    ----------
    type : ``RelationshipType``
        The relationship's type.
    user : ``User`` or ``Client`` instance
        The target user of the relationship.
    """
    __slots__ = ('type', 'user',)
    def __init__(self, client, data, user_id):
        """
        Creates a relationship instance with the given parameters.
        
        Parameters
        ----------
        client : ``Client``
            The client, who's relationship is created.
        data : `dict` of (`str`, `Any`)
            Relationship data.
        user_id : `int`
            The relationship's target user's id.
        """
        self.user = create_partial_user(user_id)
        self.type = RelationshipType.get(data['type'])
        client.relationships[user_id] = self
    
    def __repr__(self):
        """Returns the representation of the relationship."""
        return f'<{self.__class__.__name__} {self.type.name} user={self.user.full_name!r}>'

class Unknown(DiscordEntity):
    """
    Represents a not found object when creating an ``AuditLog``.
    
    Attributes
    ----------
    id : `int`
        The entity's unique identifier number.
    name : `str`
        The entity's name.
    type : `str`
        The entity's respective type's respective name.
        
        Can be one of:
        - `'Channel'`
        - `'Emoji'`
        - `'Integration'`
        - `'Invite'`
        - `'Message'`
        - `'Role'`
        - `'User'`
        - `'Webhook'`
    """
    __slots__ = ('name', 'type', )
    
    def __init__(self, type_, id_, name=None):
        """
        Creates a new ``Unknown`` object from the given parameters.
        
        Parameters
        ----------
        type_ : `str`
            The entity's respective type's respective name.
        id_ : `int`
            The entity's unique identifier number.
        name : `str`, Optional
            The name of the entity if applicable. If not, `type_` will be used as name instead.
        """
        self.type = type_
        self.id = id_
        if name is None:
            name = type_
        self.name = name
    
    def __repr__(self):
        """Returns the representation of the entity."""
        return f'<{self.__class__.__name__} type={self.type} id={self.id} name={self.name!r}>'
    
    def __gt__(self, other):
        """Returns whether this entity's respective type matches with the other's and it's id is greater than the
        other's."""
        if type(other) is type(self):
            if self.type != other.type:
                return NotImplemented
        elif isinstance(other, DiscordEntity):
            if self.type not in other.__class__.__name__:
                return NotImplemented
        else:
            return NotImplemented
        
        return (self.id > other.id)
    
    def __ge__(self ,other):
        """Returns whether this entity's respective type matches with the other's and it's id is greater or equal to
        the other's."""
        if type(other) is type(self):
            if self.type != other.type:
                return NotImplemented
        elif isinstance(other, DiscordEntity):
            if self.type not in other.__class__.__name__:
                return NotImplemented
        else:
            return NotImplemented
        
        return (self.id >= other.id)
    
    def __eq__(self, other):
        """Returns whether this entity's respective type matches with the other's and it's id equals to the other's."""
        if type(other) is type(self):
            if self.type != other.type:
                return NotImplemented
        elif isinstance(other, DiscordEntity):
            if self.type not in other.__class__.__name__:
                return NotImplemented
        else:
            return NotImplemented
        
        return (self.id == other.id)
    
    def __ne__(self,other):
        """Returns whether this entity's respective type matches with the other's and it's id not equals to the
        other's."""
        if type(other) is type(self):
            if self.type != other.type:
                return NotImplemented
        elif isinstance(other, DiscordEntity):
            if self.type not in other.__class__.__name__:
                return NotImplemented
        else:
            return NotImplemented
        
        return (self.id != other.id)
    
    def __le__(self,other):
        """Returns whether this entity's respective type matches with the other's and it's id is less or equal to
        the other's."""
        if type(other) is type(self):
            if self.type != other.type:
                return NotImplemented
        elif isinstance(other, DiscordEntity):
            if self.type not in other.__class__.__name__:
                return NotImplemented
        else:
            return NotImplemented
        
        return (self.id <= other.id)
    
    def __lt__(self,other):
        """Returns whether this entity's respective type matches with the other's and it's id is less than the
        other's."""
        if type(other) is type(self):
            if self.type != other.type:
                return NotImplemented
        elif isinstance(other, DiscordEntity):
            if self.type not in other.__class__.__name__:
                return NotImplemented
        else:
            return NotImplemented
        
        return (self.id < other.id)

class Gift:
    """
    Represents a Discord gift.
    
    Attributes
    ----------
    code : `str`
        The code of the gift.
    uses : `int`
        The amount how much time the gift can be used.
    """
    __slots__ = ('code', 'uses', )
    def __init__(self, data):
        """
        Creates a new ``Gift`` object from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Gift data received from Discord.
        """
        self.uses = data['uses']
        self.code = data['code']

@modulize
class DISCORD_HEADERS:
    # to receive
    AUDIT_LOG_REASON = istr('X-Audit-Log-Reason')
    RATE_LIMIT_REMAINING = istr('X-RateLimit-Remaining')
    RATE_LIMIT_RESET = istr('X-RateLimit-Reset')
    RATE_LIMIT_RESET_AFTER = istr('X-RateLimit-Reset-After')
    RATE_LIMIT_LIMIT = istr('X-RateLimit-Limit')
    # to send
    RATE_LIMIT_PRECISION = istr('X-RateLimit-Precision')

def url_cutter(url):
    """
    Cuts down the given url to a more suitable length.
    
    Parameters
    ----------
    url : `str`
    
    Returns
    -------
    result : `str`
    """
    if len(url) < 50:
        return url
    
    position = url.find('/')
    
    if position == -1:
        return f'{url[:28]}...{url[-19:]}'
    
    position += 1
    if url[position] == '/':
        position += 1
        if position == len(url):
            return f'{url[:28]}...{url[-19:]}'
        
        position = url.find('/', position)
        position +=1
        if position == 0 or position == len(url):
            return f'{url[:28]}...{url[-19:]}'
    
    positions = [position]
    
    while True:
        position = url.find('/', position)
        if position == -1:
            break
        position +=1
        if position == len(url):
            break
        positions.append(position)
    
    from_start = 0
    from_end = 0
    top_limit = len(url)
    index = 0
    
    while True:
        value = positions[index]
        if value+from_end > 47:
            if from_start+from_end < 33:
                from_start = 47-from_end
                break
            else:
                index +=1
                if index == len(positions):
                    value  =0
                else:
                    value = positions[len(positions)-index]
                value = top_limit-value
                if value+from_start>47:
                    break
                else:
                    from_end = value
                    break
        from_start = value
        
        index = index+1
        value = positions[len(positions)-index]
        value = top_limit-value
        if value+from_start > 47:
            if from_start+from_end < 33:
                from_end = 47-from_start
                break
            else:
                if index == len(positions):
                    value = top_limit
                else:
                    value = positions[index]
                
                if value+from_end > 47:
                    break
                else:
                    from_start = value
                    break
        from_end = value
        
    return f'{url[:from_start]}...{url[top_limit-from_end-1:]}'

DELTA_RP = re_compile('([\+\-]?\d+)[ \t]*([a-zA-Z]+)')
TDELTA_KEYS = ('weeks', 'days', 'hours', 'minutes', 'seconds', 'microseconds')
RDELTA_KEYS = ('years', 'months', *TDELTA_KEYS)

def parse_tdelta(text):
    """
    Tries to parse out a ``timedelta`` from the inputted text.
    
    Returns
    -------
    tdelta : `None` or `datetime.timedelta`
    """
    text = text.lower()
    
    result = {}
    index = 0
    limit = len(TDELTA_KEYS)
    for amount, name in DELTA_RP.findall(text):
        if index == limit:
            break
        
        while True:
            key = TDELTA_KEYS[index]
            index += 1
            if key.startswith(name):
                result.setdefault(key, int(amount))
                break
            
            if index == limit:
                break
    
    if result:
        return timedelta(**result)

if relativedelta is None:
    parse_rdelta = None
else:
    def parse_rdelta(text):
        """
        Tries to parse out a ``relativedelta`` from the inputted text.
        
        Returns
        -------
        rdelta : `None` or `dateutil.relativedelta.relativedelta`
        """
        text = text.lower()
        
        result = {}
        index = 0
        limit = len(RDELTA_KEYS)
        for amount, name in DELTA_RP.findall(text):
            if index == limit:
                break
            
            while True:
                key = RDELTA_KEYS[index]
                index += 1
                if key.startswith(name):
                    result.setdefault(key, int(amount))
                    break
                
                if index == limit:
                    break
        
        if result:
            return relativedelta(**result)

CHANNEL_MESSAGE_RP = re_compile('(\d{7,21})-(\d{7,21})')

def parse_message_reference(text):
    """
    Tries the parse a ``Message``'s reference from the inputted text.
    
    Accepts the following formats:
    - `{message.id}`
    - `{channel.id}-{message.id}`
    - `{message.url}`
    
    Returns
    -------
    reference : `None` or `tuple` (`int`, `int`, `int`)
        On successful parsing returns a tuple of 3 elements:
        +-------------------+-------------------+-------------------------------+-------------------+---------------+
        | Respective name   | Parse-able from   | Parse-able from               | Parse-able from   | Default value |
        |                   | `{message.id}`    | `{channel.id}-{message.id}`   | `{message.url}`   |               |
        +===================+===================+===============================+=================+++===============+
        | `guild.id`        | False             | False                         | True              | `0`           |
        +-------------------+-------------------+-------------------------------+-------------------+---------------+
        | `channel.id       | False             | True                          | True              | `0`           |
        +-------------------+-------------------+-------------------------------+-------------------+---------------+
        | `message.id`      | True              | True                          | True              | N/A           |
        +-------------------+-------------------+-------------------------------+-------------------+---------------+
    """
    parsed = ID_RP.fullmatch(text)
    if (parsed is not None):
        message_id = int(parsed.group(1))
        
        guild_id = 0
        channel_id = 0
    else:
        parsed = CHANNEL_MESSAGE_RP.fullmatch(text)
        if (parsed is not None):
            channel_id, message_id = parsed.groups()
            channel_id = int(channel_id)
            message_id = int(message_id)
            
            guild_id = 0
        else:
            parsed = MESSAGE_JUMP_URL_RP.fullmatch(text)
            if (parsed is not None):
                guild_id, channel_id, message_id = parsed.groups()
                if guild_id == '@me':
                    guild_id = 0
                else:
                    guild_id = int(guild_id)
                channel_id = int(channel_id)
                message_id = int(message_id)
            else:
                return None
    
    return guild_id, channel_id, message_id


def sanitize_mentions(content, guild=None):
    """
    Sanitizes the given content, removing the mentions from it.
    
    Parameters
    ----------
    content : `str`
        The content to validate.
    guild : `None` or ``Guild``, Optional
        Respective context to look up guild specific names of entities.
    
    Returns
    -------
    content : `str`
    """
    transformations = {
        '@everyone':'@\u200beveryone',
        '@here':'@\u200bhere',
            }
    
    for id_ in USER_MENTION_RP.findall(content):
        id_ = int(id_)
        user = USERS.get(id_, None)
        if (user is None) or user.partial:
            sanitized_mention = '@invalid-user'
        else:
            sanitized_mention = '@'+user.name_at(guild)
        
        transformations[f'<@{id_}>'] = sanitized_mention
        transformations[f'<@!{id_}>'] = sanitized_mention
        
    for id_ in CHANNEL_MENTION_RP.findall(content):
        id_ = int(id_)
        channel = CHANNELS.get(id_, None)
        if (channel is None) or channel.partial:
            sanitized_mention = '@deleted channel'
        else:
            sanitized_mention = '#'+channel.name
        
        transformations[f'<#{id_}>'] = sanitized_mention
    
    for id_ in ROLE_MENTION_RP.findall(content):
        id_ = int(id_)
        role = ROLES.get(id_, None)
        if (role is None) or role.partial:
            sanitized_mention = '@deleted role'
        else:
            sanitized_mention = '@'+role.name
        
        transformations[f'<@&{id_}>'] = sanitized_mention
    
    return re_compile('|'.join(transformations)).sub(
        lambda mention: transformations[mention.group(0)], content)



# -*- coding: utf-8 -*-
class PayloadError(Exception):
    """
    Raised when http payload processing fails.
    """
    pass


class InvalidHandshake(Exception):
    """
    Raised when websocket handshake fails.
    """
    pass


class HttpProcessingError(Exception):
    """
    Base class for http content specific errors.
    
    Attributes
    ----------
    code : `int`
        Http error code. Defaults to `0`.
    message : `str`
        Error message. Defaults to empty string.
    headers : `None` or ``imultidict`` of (`str`, `str`) items
        Respective headers.
    """
    def __init__(self, code=0, message='', headers=None):
        self.code = code
        self.headers = headers
        self.message = message
        
        Exception.__init__(self, f'HTTP {self.code}, message={message!r}, headers={self.headers!r}')


class AbortHandshake(HttpProcessingError, InvalidHandshake):
    """
    Raised when websocket handshake is aborted on server side.
    
    Attributes
    ----------
    code : `int`
        Http error code. Defaults to `0`.
    message : `str`
        Error message. Defaults to empty string.
    headers : `None` or ``imultidict`` of (`str`, `str`) items
        Respective headers.
    """


class ProxyError(HttpProcessingError):
    """
    Raised when a proxy request responds with status other than "200 OK".
    
    Attributes
    ----------
    code : `int`
        Http error code. Defaults to `0`.
    message : `str`
        Error message. Defaults to empty string.
    headers : `None` or ``imultidict`` of (`str`, `str`) items
        Respective headers.
    """
    pass


class InvalidOrigin(InvalidHandshake):
    """
    Raised when a websocket handshake received invalid origin header.
    """
    pass


class InvalidUpgrade(InvalidHandshake):
    """
    Raised when a websocket was not correctly upgraded.
    """
    pass


class ContentEncodingError(HttpProcessingError, PayloadError):
    """
    Raised when http content decoding fails.
    
    Attributes
    ----------
    code : `int`
        Http error code. Defaults to `0`.
    message : `str`
        Error message. Defaults to empty string.
    headers : `None` or ``imultidict`` of (`str`, `str`) items
        Respective headers.
    """
    def __init__(self, message='Bad Request', headers=None):
        HttpProcessingError.__init__(self, 400, message, headers)


class ConnectionClosed(Exception):
    """
    Connection closed exception raised when a websocket is closed.
    
    Attributes
    ----------
    code : `int`
        Websocket close code.
    exception : `None` or `BaseException`
        Source exception if applicable.
    reason : `None or `str`
        Websocket close reason if any.
    
    Class Attributes
    ----------------
    _close_reasons : `dict` of (`int`, `str`) items
        Predefined close reasons used if the respective close frame did not include one.
    """
    _close_reasons = {
        1000: 'OK',
        1001: 'going away',
        1002: 'protocol error',
        1003: 'unsupported type',
        1004: '`reserved`',
        1005: 'no status code [internal]',
        1006: 'connection closed abnormally [internal]',
        1007: 'invalid data',
        1008: 'policy violation',
        1009: 'message too big',
        1010: 'extension required',
        1011: 'unexpected error',
        1013: 'Try again later',
        1014: 'Bad gateway',
        1015: 'TLS failure [internal]',
            }

    @classmethod
    def _get_close_reason(cls, code):
        """
        Classmethod to get close for any websocket close code.
        
        Parameters
        ----------
        code : `int`
            Websocket close code.
        
        Returns
        -------
        reason : `str`
        """
        try:
            return cls._close_reasons[code]
        except KeyError:
            pass
        
        if code < 1000:
            return '`unused`'
        if code < 2000:
            return '`reserved`'
        if code < 3000:
            return '`reserved for extensions`'
        if code < 4000:
            return '`registered`'
        if code < 5000:
            return '`private use`'
        
        return '`unknown`'
    
    def __init__(self, code, exception, reason=None):
        """
        Creates a new ``ConnectionClosed`` exception from the given parameters.
        
        Parameters
        ----------
        code : `int`
            The websocket close code.
        exception : `None` or `BaseException`
            Source exception if applicable.
        reason : `None or `str`, Optional
            Websocket close reason if any. Defaults to `None`.
        """
        self.code = code
        self.exception = exception
        self._reason = reason
        Exception.__init__(self)
    
    @property
    def reason(self):
        """
        Returns the websocket close reason.
        
        Returns
        -------
        reason : `str`
        """
        reason = self._reason
        if (reason is None):
            reason = self._get_close_reason(self.code)
        
        return reason
    
    def __repr__(self):
        """Returns the exception's representation."""
        return f'<{self.__class__.__name__}, code={self.code!r}, reason={self.reason!r}, exception={self.exception!r}>'
    
    __str__ = __repr__


class WebSocketProtocolError(Exception):
    """
    Exception raised by websocket when receiving invalid payload.
    """
    pass



# -*- coding: utf-8 -*-
__all__ = ('Closer', )

from ...backend.futures import CancelledError
from ...backend.utils import copy_docs
from ...discord.emoji import BUILTIN_EMOJIS
from ...discord.parsers import InteractionEvent
from ...discord.message import Message
from ...discord import ChannelTextBase
from ...discord.exceptions import DiscordException, ERROR_CODES

from .bases import GUI_STATE_READY, PaginationBase
from .utils import Timeouter

class Closer(PaginationBase):
    """
    Familiar to ``Pagination``, but can be used if the given content has only 1 page, so only an `x` would show up.
    
    Picks up on reaction additions by any users.
    
    Attributes
    ----------
    _canceller : `None` or `function`
        The function called when the ``Closer`` is cancelled or when it expires. This is a onetime use and after
        it was used, is set as `None`.
    
    _task_flag : `int`
        A flag to store the state of the ``Closer``.
        
        Possible values:
        +---------------------------+-------+-----------------------------------------------------------------------+
        | Respective name           | Value | Description                                                           |
        +===========================+=======+=======================================================================+
        | GUI_STATE_READY           | 0     | The closer does nothing, is ready to be used.                         |
        +---------------------------+-------+-----------------------------------------------------------------------+
        | GUI_STATE_SWITCHING_PAGE  | 1     | The closer is currently changing it's page.                           |
        +---------------------------+-------+-----------------------------------------------------------------------+
        | GUI_STATE_CANCELLING      | 2     | The pagination is currently changing it's page, but it was cancelled  |
        |                           |       | meanwhile.                                                            |
        +---------------------------+-------+-----------------------------------------------------------------------+
        | GUI_STATE_CANCELLED       | 3     | The pagination is, or is being cancelled right now.                   |
        +---------------------------+-------+-----------------------------------------------------------------------+
        | GUI_STATE_SWITCHING_CTX   | 4     | The closer is switching context. Not used by the default class,       |
        |                           |       | but expected.                                                         |
        +---------------------------+-------+-----------------------------------------------------------------------+
    
    _timeouter : `None` or ``Timeouter``
        Executes the timing out feature on the ``Closer``.
    
    channel : ``ChannelTextBase`` instance
        The channel where the ``Closer`` is executed.
    
    client : ``Client`` of ``Embed`` (or any compatible)
        The client who executes the ``Closer``.
    
    message : `None` or ``Message``
        The message on what the ``Closer`` is executed.
    
    check : `None` or `callable`
        A callable what decides whether the ``Closer`` should process a received reaction event. Defaults to
        `None`.
        
        Should accept the following parameters:
        +-----------+---------------------------------------------------+
        | Name      | Type                                              |
        +===========+===================================================+
        | event     | ``ReactionAddEvent`` or ``ReactionDeleteEvent``   |
        +-----------+---------------------------------------------------+
        
        Note, that ``ReactionDeleteEvent`` is only given, when the client has no `manage_messages` permission.
        
        Should return the following values:
        +-------------------+-----------+
        | Name              | Type      |
        +===================+===========+
        | should_process    | `bool`    |
        +-------------------+-----------+
    
    Class Attributes
    ----------------
    CANCEL : ``Emoji`` = `BUILTIN_EMOJIS['x']`
        The emoji used to cancel the ``Closer``.
    """
    CANCEL = BUILTIN_EMOJIS['x']
    
    __slots__ = ('check',)
    
    async def __new__(cls, client, channel, content, *, timeout=240., message=None, check=None):
        """
        Creates a new pagination with the given parameters.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client who will execute the ``Closer``.
        channel : ``ChannelTextBase`` instance, ``Message``, ``InteractionEvent``
            The channel where the ``Closer`` will be executed.  Pass it as a ``Message`` instance to send a reply.
        
            If given as ``InteractionEvent``, then will acknowledge it and create a new message with it as well.
            Although will not acknowledge it if `message` is given.
        
        content : ``Any`
            The displayed content.
        timeout : `float`, Optional (Keyword Only)
            The timeout of the ``Closer`` in seconds. Defaults to `240.0`.
        message : `None` or ``Message``, Optional (Keyword Only)
            The message on what the ``Closer`` will be executed. If not given a new message will be created.
            Defaults to `None`.
        check : `None` or `callable`, Optional (Keyword Only)
            A callable what decides whether the ``Closer`` should process a received reaction event. Defaults to
            `None`.
            
            Should accept the following parameters:
            +-----------+---------------------------------------------------+
            | Name      | Type                                              |
            +===========+===================================================+
            | event     | ``ReactionAddEvent`` or ``ReactionDeleteEvent``   |
            +-----------+---------------------------------------------------+
            
            Note, that ``ReactionDeleteEvent`` is only given, when the client has no `manage_messages` permission.
            
            Should return the following values:
            +-------------------+-----------+
            | Name              | Type      |
            +===================+===========+
            | should_process    | `bool`    |
            +-------------------+-----------+
        
        Returns
        -------
        self : `None` or ``Closer``
        
        Raises
        ------
        TypeError
            `channel`'s type is incorrect.
        """
        if isinstance(channel, ChannelTextBase):
            target_channel = channel
            received_interaction = False
        elif isinstance(channel, Message):
            target_channel = channel.channel
            received_interaction = False
        elif isinstance(channel, InteractionEvent):
            target_channel = channel.channel
            received_interaction = True
        else:
            raise TypeError(f'`channel` can be given only as `{ChannelTextBase.__name__}`, `{Message.__name__}` '
                f'of as {InteractionEvent.__name__} instance, got {channel.__class__.__name__}.')
        
        self = object.__new__(cls)
        self.check = check
        self.client = client
        self.channel = target_channel
        self._canceller = cls._canceller_function
        self._task_flag = GUI_STATE_READY
        self.message = message
        self._timeouter = None
        
        try:
            if message is None:
                if received_interaction:
                    if not channel.is_acknowledged():
                        await client.interaction_response_message_create(channel)
                    
                    message = await client.interaction_followup_message_create(channel, content)
                else:
                    message = await client.message_create(channel, content)
                self.message = message
            else:
                await client.message_edit(message, content)
        except BaseException as err:
            self.cancel(err)
            if isinstance(err, ConnectionError):
                return self
            
            if isinstance(err, DiscordException):
                if err.code in (
                        ERROR_CODES.unknown_message, # message deleted
                        ERROR_CODES.unknown_channel, # message's channel deleted
                        ERROR_CODES.invalid_access, # client removed
                        ERROR_CODES.invalid_permissions, # permissions changed meanwhile
                        ERROR_CODES.cannot_message_user, # user has dm-s disallowed
                            ):
                    return self
            
            raise
        
        
        if not target_channel.cached_permissions_for(client).can_add_reactions:
            self.cancel(PermissionError())
            return self
        
        try:
            await client.reaction_add(message, self.CANCEL)
        except BaseException as err:
            self.cancel(err)
            if isinstance(err, ConnectionError):
                return self
            
            if isinstance(err, DiscordException):
                if err.code in (
                        ERROR_CODES.unknown_message, # message deleted
                        ERROR_CODES.unknown_channel, # message's channel deleted
                        ERROR_CODES.max_reactions, # reached reaction 20, some1 is trolling us.
                        ERROR_CODES.invalid_access, # client removed
                        ERROR_CODES.invalid_permissions, # permissions changed meanwhile
                            ):
                    return self
            
            raise
        
        
        self._timeouter = Timeouter(self, timeout=timeout)
        client.events.reaction_add.append(message, self)
        return self
    
    
    @copy_docs(PaginationBase.__call__)
    async def __call__(self, client, event):
        if event.user.is_bot:
            return
        
        if (event.emoji is not self.CANCEL):
            return
        
        task_flag = self._task_flag
        if task_flag != GUI_STATE_READY:
            # ignore GUI_STATE_SWITCHING_PAGE and GUI_STATE_CANCELLED and GUI_STATE_SWITCHING_CTX
            return
        
        check = self.check
        if (check is not None):
            try:
                should_continue = check(event)
            except BaseException as err:
                await client.events.error(client, f'{self!r}.__call__', err)
                return
            
            if not should_continue:
                return
        
        self.cancel(CancelledError())



# -*- coding: utf-8 -*-
__all__ = ('AsyncIO', 'ReuAsyncIO', 'ReuBytesIO', )

import os
from io import BytesIO
from threading import current_thread

from .executor import ExecutorThread
from .utils import alchemy_incendiary
from .event_loop import EventThread

OPERATION_WRITE = 0
OPERATION_READ = 1

IO_CLOSED_OR_DETACHED = 'I/O operation on closed or on a detached file.'

class ReuBytesIO(BytesIO):
    """
    Reusable bytes io, what seeks the cursor at `0`, when calling ``.close``. Use ``.real_close`` to close it for real,
    or use it with the `with` syntax.
    
    Should be used instead of `BytesIO` when sending it with http requests. Since the buffer is closed when a request
    is sent, at the case of received errors repeat could not be executed.
    
    Attributes
    ----------
    _last_op : `int`
        The last called operation of the buffer.
        
        If switching after reading to writing the buffer seeks to `0`.
        
        Can be set as `1` of the following:
        
        +-------------------+-------+
        | Respective name   | Value |
        +===================+=======+
        | OPERATION_WRITE   | 0     |
        +-------------------+-------+
        | OPERATION_READ    | 1     |
        +-------------------+-------+
    _size : `int`
        The position, till the buffer was written by the last reading session and can be read back.
    """
    __slots__ = ('_last_op', '_size')
    
    def __init__(self,):
        """
        Initializes the buffer.
        """
        self._size = 0
        self._last_op = OPERATION_WRITE
    
    def write(self, data):
        """
        Writes the given data to the buffer.
        
        If the buffer is in `read` mode, in seeks back to `0`
        
        Parameters
        ----------
        data : `bytes-like`
            The data to write.
        """
        if self._last_op != OPERATION_WRITE:
            BytesIO.seek(self, 0)
            self._size = 0
            self._last_op = OPERATION_WRITE
        
        amount = BytesIO.write(self, data)
        self._size += amount
        return amount
    
    def read(self, amount=None):
        """
        Reads the given amount of data from the buffer.
        
        Parameters
        ----------
        amount : `None` or `int`, Optional
            The amount of data to read from the buffer. If given as `None`, so by the default, then reads all the data
            out from it.
        
        Returns
        -------
        data : `bytes`
            The red data.
        """
        self._last_op = OPERATION_READ
        
        if amount is None:
            amount = self._size-self.tell()
        else:
            readable = self._size-self.tell()
            if amount > readable:
                amount = readable
        
        return BytesIO.read(self, amount)
    
    def close(self):
        """
        Seeks back to position `0`.
        """
        self.seek(0)
    
    def __len__(self):
        """
        Returns the length till the buffer was written in it's last write session.
        
        Returns
        -------
        length : `int`
        """
        return self._size
    
    def seek(self, offset, whence=os.SEEK_SET):
        """
        Changes the stream position by the given offset. Offset is interpreted relative to the position indicated
        by `whence`.
        
        Parameters
        ----------
        offset : `int`
            Position to move the cursor to.
        whence : `int`, Optional
            How the given `offset` should be interpreted. Default value for whence is `os.SEEK_SET`.
            
            Can be given as:
            
            +-------------------+-------+-----------------------------------------------------------+
            | Respective name   | Value | Description                                               |
            +===================+=======+===========================================================+
            | os.SEEK_SET       | 0     | Start of the stream. Offset should be zero or positive.   |
            +-------------------+-------+-----------------------------------------------------------+
            | os.SEEK_CUR       | 1     | Current stream position. Offset may be negative.          |
            +-------------------+-------+-----------------------------------------------------------+
            | os.SEEK_END       | 2     | End of the stream. Offset is usually negative.            |
            +-------------------+-------+-----------------------------------------------------------+
            
        Returns
        -------
        position : `int`
            The new absolute position.
        """
        if whence == os.SEEK_END:
            return self._size
        
        value = BytesIO.seek(self, offset, whence)
        if value > self._size:
            self._size = value
            
        return value
    
    def real_close(self):
        """
        Closes the buffer.
        """
        BytesIO.close(self)


def get_executor():
    """
    Gets executor from the current event thread if applicable. If not, starts a new one.
    
    Returns
    -------
    executor : ``ExecutorThread`` or ``ClaimedExecutor``
    """
    loop = current_thread()
    if isinstance(loop, EventThread):
        executor = loop.claim_executor()
    else:
        executor = ExecutorThread()
    
    return executor


class _AsyncIOIterator:
    """
    Asynchronous iterator for ``AsyncIO`` objects.
    
    Attributes
    ----------
    _wrapped : ``AsyncIO``
        The ``AsyncIO`` to asynchronous iterate over.
    """
    __slots__ = ('_wrapped',)
    
    def __init__(self, wrapped):
        """
        Creates a mew ``_AsyncIOIterator`` instance with the given parameter.
        
        Parameters
        ----------
        wrapped : ``AsyncIO``
            The ``AsyncIO`` to asynchronous iterate over.
        """
        self._wrapped = wrapped
    
    def __aiter__(self):
        """Asynchronous iterating an ``_AsyncIOIterator`` returns itself."""
        return self
    
    async def __anext__(self):
        """
        Reads a line from the respective file.
        
        This method is a coroutine.
        """
        result = await self._wrapped.readline()
        if result:
            return result
        
        raise StopAsyncIteration


class AsyncIO:
    """
    Asynchronous file-io, what uses an executor for it.
    
    Attributes
    ----------
    _executor : `None`, ``ExecutorThread`` or ``ClaimedExecutor``
        The executor what executes the io waiting. Set as `None` if the respective `file-io` is closed.
    _io : `file-io`
        The wrapped `file-io`.
    """
    __slots__ = ('_executor', '_io',)
    
    async def __new__(cls, *args, **kwargs):
        """
        Creates a new ``AsyncIO`` instance with the given parameters.
        
        This method is a coroutine.
        
        Parameters
        ----------
        *args : arguments
            Arguments to use when opening the respective file.
        **kwargs : keyword arguments
            Keyword arguments to use when opening the respective file.
        
        Notes
        -----
        Check `https://docs.python.org/3/library/functions.html#open` for parameters.
        """
        self = object.__new__(cls)
        self._executor = executor = get_executor()
        self._io = await executor.execute(alchemy_incendiary(open, args, kwargs))
        return self
    
    @classmethod
    def wrap(cls, io):
        """
        Wraps an already open file.
        
        Parameters
        ----------
        io : `file-io`
            The opened `file-io` to wrap.
        
        Returns
        -------
        self : ``AsyncIO``
        """
        self = object.__new__(cls)
        self._io = io
        if io.closed:
            executor = None
        else:
            executor = get_executor()
        self._executor = executor
        return self
    
    @property
    def buffer(self):
        """
        Returns the buffer of the underlying stream.
        
        Returns
        -------
        buffer : ``_io.BufferedWriter`` or ``_io.BufferedReader``
        """
        return self._io.buffer
    
    def __del__(self):
        """Releases the executor and closes the wrapped `file-io` if not yet done."""
        executor = self._executor
        if executor is None:
            return
        
        self._io.close()
        executor.release()
        self._executor = None
    
    close = __del__
    
    @property
    def closed(self):
        """
        Returns whether the ``AsyncIO`` instance is closed.
        
        Returns
        -------
        closed : `bool`
        """
        return (self._executor is None)
    
    async def detach(self):
        """
        Separates the wrapped `file-io`'s underlying raw stream from the buffer and return it.
        
        After detaching the ``AsyncIO`` becomes unusable.
        
        This method is a coroutine.
        
        Returns
        -------
        buffer : ``_io.BufferedWriter`` or ``_io.BufferedReader``
        
        Raises
        ------
        ValueError
            I/O operation on closed or on a detached file.
        """
        executor = self._executor
        if executor is None:
            raise ValueError(IO_CLOSED_OR_DETACHED)
        
        raw = await executor.execute(self._io.detach)
        self._executor = None
        return raw
    
    async def detach_to_self(self):
        """
        Separates the wrapped `file-io`'s underlying raw stream from the buffer and attaches it to the ``AsyncIO``
        
        This method is a coroutine.
        
        Raises
        ------
        ValueError
            I/O operation on closed or on a detached file.
        """
        executor = self._executor
        if executor is None:
            raise ValueError(IO_CLOSED_OR_DETACHED)
        self._io = await executor.execute(self._io.detach)
    
    @property
    def encoding(self):
        """
        Returns the io-s encoding.
        
        Returns
        -------
        encoding : `str`
        """
        return self._io.encoding
    
    @property
    def errors(self):
        """
        Returns the error settings of the io.
        
        Returns
        -------
        errors : `str`
        """
        return self._io.errors
    
    def fileno(self):
        """
        Returns the underlying file descriptor.
        
        Returns
        -------
        fd : `int`
        """
        return self._io.fileno()
    
    async def flush(self):
        """
        Force bytes held in the buffer into the raw stream.
        
        This method is a coroutine.
        
        Raises
        ------
        BlockingIOError
            If the raw stream blocks.
        ValueError
            I/O operation on closed or on a detached file.
        """
        executor = self._executor
        if executor is None:
            raise ValueError(IO_CLOSED_OR_DETACHED)
        
        return await executor.execute(self._io.flush)
        
    def isatty(self):
        """
        Returns whether the stream is interactive.
        
        Returns
        -------
        isatty : `bool`
        """
        return self._io.isatty()
    
    @property
    def line_buffering(self):
        """
        Returns whether line buffering is enabled.
        
        Returns
        -------
        line_buffering : `bool`
        """
        return self._io.line_buffering
    
    @property
    def mode(self):
        """
        Returns the `mode` given to the constructor.
        
        Returns
        -------
        mode : `str`
        """
        return self._io.mode
    
    @property
    def name(self):
        """
        The file's name.
        
        Returns
        -------
        name : `str`
        """
        return self._io.name
    
    @property
    def newlines(self):
        """
        Indicating the newlines translated so far. Depending on the implementation and the initial constructor flags,
        this may not be available.
        
        Returns
        -------
        newlines : `None`, `str` or `tuple` of `str`
        """
        return self._io.newlines
    
    async def read(self, size=-1):
        """
        Reads from the underlying stream.
        
        This method is a coroutine.
        
        Parameters
        ----------
        size : `int`
            The upper limit to read from the underlying stream. If eof is received meanwhile reading, the already read
            data is returned.
            
            If given as negative `int`, then all the data till eof is read from the underlying stream.
        
        Returns
        -------
        data : `bytes` or `str`
            The read data.
        
        Raises
        ------
        BlockingIOError
            Might be raised if the underlying stream is in blocking mode.
        ValueError
            I/O operation on closed or on a detached file.
        """
        executor = self._executor
        if executor is None:
            raise ValueError(IO_CLOSED_OR_DETACHED)
        
        return await executor.execute(alchemy_incendiary(self._io.read, (size,),))
    
    def read1(self, *args):
        """
        Reads from the underlying stream.
        
        This method is a coroutine.
        
        Can be used for implementing your own buffering over ``.read``.
        
        Parameters
        ----------
        size : `int`
            The upper limit to read from the underlying stream. If eof is received meanwhile reading, the already read
            data is returned.
            
            If given as negative `int`, then all the data till eof is read from the underlying stream.
        
        Returns
        -------
        data : `bytes` or `str`
            The read data.
        
        Raises
        ------
        ValueError
            I/O operation on closed or on a detached file.
        """
        executor = self._executor
        if executor is None:
            raise ValueError(IO_CLOSED_OR_DETACHED)
        
        return executor.execute(alchemy_incendiary(self._io.read1, args,))
    
    @property
    def readable(self):
        """
        Returns whether the underlying stream is readable.
        
        Returns
        -------
        readable : `bool`
        """
        return self._io.readable
    
    async def readinto(self, b):
        """
        Read bytes into a pre-allocated writable `bytes-like` object `b`, and return the number of bytes read.
        
        This method is a coroutine.
        
        Parameters
        ----------
        b : `bytes-like`
            Writable bytes like to write into.
        
        Raises
        ------
        BlockingIOError
            Might be raised if the underlying stream is in blocking mode.
        ValueError
            I/O operation on closed or on a detached file.
        """
        executor = self._executor
        if executor is None:
            raise ValueError(IO_CLOSED_OR_DETACHED)
        
        return await executor.execute(alchemy_incendiary(self._io.readinto, (b,),))
    
    async def readinto1(self, b):
        """
        Read bytes into a pre-allocated writable `bytes-like` object `b`, and return the number of bytes read.
        
        This method is a coroutine.
        
        Can be used for implementing your own buffering over ``.readinto``.
        
        Parameters
        ----------
        b : `bytes-like`
            Writable bytes like to write into.
        
        Raises
        ------
        BlockingIOError
            Might be raised if the underlying stream is in blocking mode.
        ValueError
            I/O operation on closed or on a detached file.
        """
        executor = self._executor
        if executor is None:
            raise ValueError(IO_CLOSED_OR_DETACHED)
        
        return await executor.execute(alchemy_incendiary(self._io.readinto1, (b,),))
    
    async def readline(self, size=-1):
        """
        Read until newline or EOF and return a single `str`. If the buffer is at EOF, returns an empty string.
        
        This method is a coroutine.
        
        Parameters
        ----------
        size : `int`
            The upper limit to read from the underlying stream. If eof is received meanwhile reading, the already read data is
            returned.
            
            If given as negative `int`, then all the data till linebreak or eof is read from the underlying stream.
        
        Returns
        -------
        line : `bytes` or `str`
            The red line.
        
        Raises
        ------
        BlockingIOError
            Might be raised if the underlying stream is in blocking mode.
        ValueError
            I/O operation on closed or on a detached file.
        """
        executor = self._executor
        if executor is None:
            raise ValueError(IO_CLOSED_OR_DETACHED)
        
        return await executor.execute(alchemy_incendiary(self._io.readline, (size,),))
    
    async def readlines(self, hint=-1):
        """
        Read and return a list of lines from the stream.
        
        This method is a coroutine.
        
        Parameters
        ----------
        hint : `int`
            The upper limit if lines to read.
            
            If given as negative `int`, then all the data till eof is read from the underlying stream.
        
        Raises
        ------
        BlockingIOError
            Might be raised if the underlying stream is in blocking mode.
        ValueError
            I/O operation on closed or on a detached file.
        """
        executor = self._executor
        if executor is None:
            raise ValueError(IO_CLOSED_OR_DETACHED)
        
        io = self._io
        return await executor.execute(alchemy_incendiary(io.__class__.readlines, (io, hint,),))
    
    async def seek(self, offset, whence=os.SEEK_SET):
        """
        Changes the stream position by the given offset. Offset is interpreted relative to the position indicated
        by `whence`.
        
        This method is a coroutine.
        
        Parameters
        ----------
        offset : `int`
            Position to move the cursor to.
        whence : `int`, Optional
            How the given `offset` should be interpreted. Default value for whence is `os.SEEK_SET`.
            
            Can be given as:
            
            +-------------------+-------+-----------------------------------------------------------+
            | Respective name   | Value | Description                                               |
            +===================+=======+===========================================================+
            | os.SEEK_SET       | 0     | Start of the stream. Offset should be zero or positive.   |
            +-------------------+-------+-----------------------------------------------------------+
            | os.SEEK_CUR       | 1     | Current stream position. Offset may be negative.          |
            +-------------------+-------+-----------------------------------------------------------+
            | os.SEEK_END       | 2     | End of the stream. Offset is usually negative.            |
            +-------------------+-------+-----------------------------------------------------------+
            
        Returns
        -------
        position : `int`
            The new absolute position.
        
        Raises
        ------
        ValueError
            I/O operation on closed or on a detached file.
        """
        executor = self._executor
        if executor is None:
            raise ValueError(IO_CLOSED_OR_DETACHED)
        
        io = self._io
        return await executor.execute(alchemy_incendiary(io.__class__.seek, (io, offset, whence),))
    
    def seekable(self):
        """
        Returns whether the stream supports random access. If not, then ``.seek``, ``.tell`` and ``.truncate`` will
        raise `OSError`.
        
        Returns
        -------
        seekable : `bool`
        """
        return self._io.seekable()
    
    async def tell(self):
        """
        Return the current stream position.
        
        This method is a coroutine
        
        Returns
        -------
        position : `int`
        
        Raises
        ------
        ValueError
            I/O operation on closed or on a detached file.
        OSError
            If the underlying stream is not seekable.
        """
        executor = self._executor
        if executor is None:
            raise ValueError(IO_CLOSED_OR_DETACHED)
        
        return await executor.execute(self._io.tell)
    
    async def truncate(self, size=None):
        """
        Resize the stream to the given size in bytes (or the current position if size is not specified). The current
        stream position isn’t changed. This resizing can extend or reduce the current file size. In case of extension,
        the contents of the new file area depend on the platform (on most systems, additional bytes are zero-filled).
        
        This method is a coroutine.
        
        Parameters
        ----------
        size : `None` or `int`, Optional
            The stream's new size in bytes.
        
        Returns
        -------
        size : `int`
            The new file size.
        
        Raises
        ------
        ValueError
            I/O operation on closed or on a detached file.
        OSError
            If the underlying stream is not seekable or writeable.
        """
        executor = self._executor
        if executor is None:
            raise ValueError(IO_CLOSED_OR_DETACHED)
        
        io = self._io
        return await executor.execute(alchemy_incendiary(io.__class__.truncate, (io, size,),))
    
    def writable(self):
        """
        Returns whether the stream supports writing. If not, then, then ``.write``, ``.writelines``  and ``.truncate``
        will raise `OSError`.
        
        Returns
        -------
        writable : `bool`
        """
        return self._io.writable()
    
    async def write(self, b):
        """
        Write the given object, to the underlying raw stream, and return the number of bytes written.
        
        Depending on the implementation the written's data's amount might be less as the given object's length.
        
        This method is a coroutine.
        
        Parameters
        ----------
        b : `bytes-like` or `str`
            The object to write into the underlying stream.
        
        Returns
        -------
        written : `int`
            The amount of bytes written to the underlying stream.
        
        Raises
        ------
        ValueError
            I/O operation on closed or on a detached file.
        OSError
            If the underlying stream is not writeable.
        """
        executor = self._executor
        if executor is None:
            raise ValueError(IO_CLOSED_OR_DETACHED)
        
        io = self._io
        return await executor.execute(alchemy_incendiary(io.__class__.write, (io, b,),))
    
    async def writelines(self, lines):
        """
        Write a list of lines to the stream. Line separators are not added, so it is usual for each of the lines
        provided to have a line separator at the end.
        
        This method is a coroutine.
        
        Parameters
        ----------
        b : `list` of (`bytes-like` or `str`)
            The lines to write into the underlying stream.
        
        Raises
        ------
        ValueError
            I/O operation on closed or on a detached file.
        OSError
            If the underlying stream is not writeable.
        """
        executor = self._executor
        if executor is None:
            raise ValueError(IO_CLOSED_OR_DETACHED)
        
        io = self._io
        return await executor.execute(alchemy_incendiary(io.__class__.writelines, (io, lines,),))
    
    def __repr__(self):
        """Returns the io's representation"""
        result = [
            '<',
            self.__class__.__name__,
            ' io=',
            repr(self._io),
                ]
        
        executor = self._executor
        if executor is None:
            result.append(' closed')
        else:
            result.append(', executor=')
            result.append(repr(executor))
        
        result.append('>')
        
        return ''.join(result)
    
    __str__ = __repr__
    
    def __enter__(self):
        """
        Enters the asynchronous io as a context manager, closing it, when exited. Raises `ValueError` if already
        closed.
        """
        if self._executor is None:
            raise ValueError('Can not enter an already closed or detached io.')
        
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        """Closes the asynchronous io."""
        self.__del__()
    
    def __aiter__(self):
        """
        Asynchronous iterates the asynchronous file-io.
        
        Returns
        -------
        iterator : ``_AsyncIOIterator``
        """
        return _AsyncIOIterator(self)


class _ReuAsyncIOIterator:
    """
    Asynchronous iterator for ``ReuAsyncIO`` objects.
    
    Attributes
    ----------
    _wrapped : ``ReuAsyncIO``
        The ``ReuAsyncIO`` to asynchronous iterate over.
    """
    __slots__ = ('_wrapped',)
    
    def __init__(self,wrapped):
        """
        Creates a mew ``_ReuAsyncIOIterator`` instance with the given parameter.
        
        Parameters
        ----------
        wrapped : ``ReuAsyncIO``
            The ``ReuAsyncIO`` to asynchronous iterate over.
        """
        self._wrapped = wrapped
    
    def __aiter__(self):
        """Asynchronous iterating an ``_ReuAsyncIOIterator`` returns itself."""
        return self
    
    async def __anext__(self):
        """Reads a line from the respective file."""
        wrapped = self._wrapped
        if wrapped._should_seek:
            await AsyncIO.seek(wrapped, 0)
            wrapped._should_seek = False
        
        result = await wrapped.readline()
        if result:
            return result
        raise StopAsyncIteration


class ReuAsyncIO(AsyncIO):
    """
    Reusable asynchronous file-io, supporting only opening in `read-bytes` mode, what seeks the cursor at `0`, when
    calling ``.close``. Use ``.real_close`` to close it for real, or use it with the `with` syntax.
    
    Should be used instead of `BytesIO` when sending it with http requests. Since the buffer is closed when a request
    is sent, at the case of received errors repeat could not be executed.
    
    Attributes
    ----------
    _executor : `None`, ``ExecutorThread`` or ``ClaimedExecutor``
        The executor what executes the io waiting. Set as `None` if the respective `file-io` is closed.
    _io : `file-io`
        The wrapped `file-io`.
    _should_seek : `bool`
        Whether the buffer was closed with ``.close`` and the next operation should seek back to `0`.
    """
    __slots__ = ('_executor', '_io', '_should_seek',)
    
    async def __new__(cls, path, mode='rb', *args, **kwargs):
        """
        Creates a new ``AsyncIO`` instance with the given parameters.
        
        This method is a coroutine.
        
        Parameters
        ----------
        path : `bytes` or `str`
            The file's path to open.
        mode : `str`
            The mode to open with the file. Defaults to `'rb'`.
        *args : arguments
            Arguments to use when opening the respective file.
        **kwargs : keyword arguments
            Keyword arguments to use when opening the respective file.
        
        Raises
        ------
        ValueError
            If `mode` is not given as `'rb'`, since ``ReuAsyncIO`` supports only `'rb'` mode.
        
        Notes
        -----
        Check `https://docs.python.org/3/library/functions.html#open` for parameters.
        """
        if mode != 'rb':
            raise ValueError(f'{cls.__name__} supports \'rb\' mode only, got {mode!r}.')
        
        self = object.__new__(cls)
        self._executor = executor = get_executor()
        self._io = await executor.execute(alchemy_incendiary(open, (path, mode, *args), kwargs))
        self._should_seek = False
        return self
    
    def close(self):
        """
        Sets the internal ``._should_seek`` slot to `True`, marking the io to seek back to the start at the next
        operation.
        """
        self._should_seek = True
    
    real_close = AsyncIO.__del__
    
    @classmethod
    def wrap(cls, io):
        """
        Wraps an already open file.
        
        Parameters
        ----------
        io : `file-io`
            The opened `file-io` to wrap.
        
        Returns
        -------
        self : ``ReuAsyncIO``
        """
        self = object.__new__(cls)
        self._io = io
        if io.closed:
            executor = None
        else:
            executor = get_executor()
        self._executor = executor
        self._should_seek = False
        return self
    
    async def detach(self):
        """
        Separates the wrapped `file-io`'s underlying raw stream from the buffer and return it.
        
        After detaching the ``ReuAsyncIO`` becomes unusable.
        
        This method is a coroutine.
        
        Returns
        -------
        buffer : ``_io.BufferedWriter`` or ``_io.BufferedReader``
        
        Raises
        ------
        ValueError
            I/O operation on closed or on a detached file.
        """
        executor = self._executor
        if executor is None:
            raise ValueError(IO_CLOSED_OR_DETACHED)
        
        raw = await executor.execute(self._io.detach)
        self._executor = None
        self._should_seek = False
        return raw
    
    @staticmethod
    def _seek_and_call(self, func, *args):
        """
        Staticmethod to seek to the respective stream's  `0` start, then call the given `func`.
        
        Parameters
        ----------
        self : ``ReuAsyncIO``
            The respective asynchronous io instance.
        func : `callable`
            The callable to call after seeking.
        *args : Arguments
            Additional arguments to call the given `func` with.
        
        Returns
        -------
        result : `Any`
            The returned value by the given `func`.
        """
        io = self._io
        io.seek(0)
        self._should_seek = False
        return func(self._io, *args)
    
    async def read(self, size=-1):
        """
        Reads from the underlying stream.
        
        This method is a coroutine.
        
        Parameters
        ----------
        size : `int`
            The upper limit to read from the underlying stream. If eof is received meanwhile reading, the already read
            data is returned.
            
            If given as negative `int`, then all the data till eof is read from the underlying stream.
        
        Returns
        -------
        data : `bytes` or `str`
            The read data.
        
        Raises
        ------
        BlockingIOError
            Might be raised if the underlying stream is in blocking mode.
        ValueError
            I/O operation on closed or on a detached file.
        """
        executor = self._executor
        if executor is None:
            raise ValueError(IO_CLOSED_OR_DETACHED)
        
        func = self._io.__class__.read
        if self._should_seek:
            task = alchemy_incendiary(self._seek_and_call, (self, func, size,),)
        else:
            task = alchemy_incendiary(func, (self._io, size,),)
        
        return await executor.execute(task)
        
    async def read1(self, *args):
        """
        Reads from the underlying stream.
        
        This method is a coroutine.
        
        Can be used for implementing your own buffering over ``.read``.
        
        Parameters
        ----------
        size : `int`
            The upper limit to read from the underlying stream. If eof is received meanwhile reading, the already read
            data is returned.
            
            If given as negative `int`, then all the data till eof is read from the underlying stream.
        
        Returns
        -------
        data : `bytes` or `str`
            The read data.
        
        Raises
        ------
        ValueError
            I/O operation on closed or on a detached file.
        """
        executor = self._executor
        if executor is None:
            raise ValueError(IO_CLOSED_OR_DETACHED)
        
        func = self._io.__class__.read1
        if self._should_seek:
            task = alchemy_incendiary(self._seek_and_call, (self, func, *args,),)
        else:
            task = alchemy_incendiary(func, (self._io, *args),)
        
        return await executor.execute(task)
    
    async def readinto(self, b):
        """
        Read bytes into a pre-allocated writable `bytes-like` object `b`, and return the number of bytes read.
        
        This method is a coroutine.
        
        Parameters
        ----------
        b : `bytes-like`
            Writable bytes like to write into.
        
        Raises
        ------
        BlockingIOError
            Might be raised if the underlying stream is in blocking mode.
        ValueError
            I/O operation on closed or on a detached file.
        """
        executor = self._executor
        if executor is None:
            raise ValueError(IO_CLOSED_OR_DETACHED)
        
        func = self._io.__class__.readinto
        if self._should_seek:
            task = alchemy_incendiary(self._seek_and_call, (self, func, b,),)
        else:
            task = alchemy_incendiary(func, (self._io, b,),)
        
        return await executor.execute(task)
    
    async def readinto1(self, b):
        """
        Read bytes into a pre-allocated writable `bytes-like` object `b`, and return the number of bytes read.
        
        This method is a coroutine.
        
        Can be used for implementing your own buffering over ``.readinto``.
        
        Parameters
        ----------
        b : `bytes-like`
            Writable bytes like to write into.
        
        Raises
        ------
        BlockingIOError
            Might be raised if the underlying stream is in blocking mode.
        ValueError
            I/O operation on closed or on a detached file.
        """
        executor = self._executor
        if executor is None:
            raise ValueError(IO_CLOSED_OR_DETACHED)
        
        func = self._io.__class__.readinto1
        if self._should_seek:
            task = alchemy_incendiary(self._seek_and_call, (self, func, b,),)
        else:
            task = alchemy_incendiary(func, (self._io, b,),)
        
        return await executor.execute(task)
    
    async def readline(self,size=-1):
        """
        Read until newline or EOF and return a single `str`. If the buffer is at EOF, returns an empty string.
        
        This method is a coroutine.
        
        Parameters
        ----------
        size : `int`
            The upper limit to read from the underlying stream. If eof is received meanwhile reading, the already read data is
            returned.
            
            If given as negative `int`, then all the data till linebreak or eof is read from the underlying stream.
        
        Returns
        -------
        line : `bytes` or `str`
            The red line.
        
        Raises
        ------
        BlockingIOError
            Might be raised if the underlying stream is in blocking mode.
        ValueError
            I/O operation on closed or on a detached file.
        """
        executor = self._executor
        if executor is None:
            raise ValueError(IO_CLOSED_OR_DETACHED)
        
        func = self._io.__class__.readline
        if self._should_seek:
            task = alchemy_incendiary(self._seek_and_call, (self, func, size,),)
        else:
            task = alchemy_incendiary(func, (self._io, size,),)
        
        return await executor.execute(task)
    
    async def readlines(self, hint=-1):
        """
        Read and return a list of lines from the stream.
        
        This method is a coroutine.
        
        Parameters
        ----------
        hint : `int`
            The upper limit if lines to read.
            
            If given as negative `int`, then all the data till eof is read from the underlying stream.
        
        Raises
        ------
        BlockingIOError
            Might be raised if the underlying stream is in blocking mode.
        ValueError
            I/O operation on closed or on a detached file.
        """
        executor = self._executor
        if executor is None:
            raise ValueError(IO_CLOSED_OR_DETACHED)
        
        func = self._io.__class__.readlines
        if self._should_seek:
            task = alchemy_incendiary(self._seek_and_call, (self, func, hint,),)
        else:
            task = alchemy_incendiary(func, (self._io, hint,),)
        
        return await executor.execute(task)
    
    async def seek(self, offset, whence=os.SEEK_SET):
        """
        Changes the stream position by the given offset. Offset is interpreted relative to the position indicated
        by `whence`.
        
        This method is a coroutine.
        
        Parameters
        ----------
        offset : `int`
            Position to move the cursor to.
        whence : `int`, Optional
            How the given `offset` should be interpreted. Default value for whence is `os.SEEK_SET`.
            
            Can be given as:
            
            +-------------------+-------+-----------------------------------------------------------+
            | Respective name   | Value | Description                                               |
            +===================+=======+===========================================================+
            | os.SEEK_SET       | 0     | Start of the stream. Offset should be zero or positive.   |
            +-------------------+-------+-----------------------------------------------------------+
            | os.SEEK_CUR       | 1     | Current stream position. Offset may be negative.          |
            +-------------------+-------+-----------------------------------------------------------+
            | os.SEEK_END       | 2     | End of the stream. Offset is usually negative.            |
            +-------------------+-------+-----------------------------------------------------------+
            
        Returns
        -------
        position : `int`
            The new absolute position.
        
        Raises
        ------
        ValueError
            I/O operation on closed or on a detached file.
        """
        executor = self._executor
        if executor is None:
            raise ValueError(IO_CLOSED_OR_DETACHED)
        
        func = self._io.__class__.seek
        if self._should_seek:
            task = alchemy_incendiary(self._seek_and_call, (self, func, offset, whence,),)
        else:
            task = alchemy_incendiary(func, (self._io, offset, whence),)
        
        return await executor.execute(task)
    
    async def tell(self):
        """
        Return the current stream position.
        
        This method is a coroutine
        
        Returns
        -------
        position : `int`
        
        Raises
        ------
        ValueError
            I/O operation on closed or on a detached file.
        OSError
            If the underlying stream is not seekable.
        """
        executor = self._executor
        if executor is None:
            raise ValueError(IO_CLOSED_OR_DETACHED)
        
        if self._should_seek:
            io = self._io
            await executor.execute(alchemy_incendiary(io.__class__.seek, (io, 0,),))
            self._should_seek = False
            return 0
        
        return await executor.execute(self._io.tell)
    
    def __aiter__(self):
        """
        Asynchronous iterates the asynchronous file-io.
        
        Returns
        -------
        iterator : ``_ReuAsyncIOIterator``
        """
        return _ReuAsyncIOIterator(self)
